# -*- coding: utf-8 -*-
import logging
import tornado.gen
import tornado.ioloop
import tornado.web
import tornado.websocket

__all__ = ('WebSocketServer',)
logger = logging.getLogger('mease.websocket_server')


class WebSocketHandler(tornado.websocket.WebSocketHandler):
    """
    Tornado websocket handler
    """
    def open(self):
        """
        Called when a client opens a websocket connection
        """
        logger.debug("Connection opened ({ip})".format(ip=self.request.remote_ip))

        # Initialize an empty client storage
        self.storage = {}

        # Call openers callbacks
        self.application._mease.call_openers(self, self.application.clients)

        # Append client to clients list
        if self not in self.application.clients:
            self.application.clients.append(self)

    def on_close(self):
        """
        Called when a client closes a websocket connection
        """
        logger.debug("Connection closed ({ip})".format(ip=self.request.remote_ip))

        # Call closer callbacks
        self.application._mease.call_closers(self, self.application.clients)

        # Remove client from clients list
        if self in self.application.clients:
            self.application.clients.remove(self)

    def on_message(self, message):
        """
        Called when a client sends a message through websocket
        """
        logger.debug("Incoming message ({ip}) : {message}".format(
            ip=self.request.remote_ip, message=message))

        # Call receiver callbacks
        self.application._mease.call_receivers(self, self.application.clients, message)

    def write_message(self, message, *args, **kwargs):
        """
        Logs message
        """
        logger.debug("Outgoing message for ({ip}) : {message}".format(
            ip=self.request.remote_ip, message=message))

        super(WebSocketHandler, self).write_message(message, *args, **kwargs)

    def send(self, *args, **kwargs):
        """
        Alias for WebSocketHandler `write_message` method
        """
        self.write_message(*args, **kwargs)


class WebSocketServer(object):
    """
    Tornado websocket server
    """
    def __init__(self, mease, port, autoreload):
        """
        Inits websocket server
        """
        self.mease = mease
        self.port = port

        # Tornado app
        self.ioloop = tornado.ioloop.IOLoop.instance()
        self.application = tornado.web.Application([
            (r'/', WebSocketHandler),
        ], debug=autoreload)

        # Initialize an empty application storage
        self.application.storage = {}

        # Initialize an empty clients list
        self.application.clients = []

        # Expose mease instance to Tornado application
        self.application._mease = self.mease

        # Connect to subscriber
        logger.debug("Connecting to backend ({backend_name})...".format(
            backend_name=self.mease.backend.name))

        self.mease.subscriber.connect()
        self.mease.subscriber.application = self.application

        # Log registered callbacks
        logger.debug("Registered callback functions :")

        logger.debug(
            "Openers : [%s]" % self._get_registry_names(self.mease.openers))
        logger.debug(
            "Closers : [%s]" % self._get_registry_names(self.mease.closers))
        logger.debug(
            "Receivers : [%s]" % self._get_registry_names(self.mease.receivers))
        logger.debug(
            "Senders : [%s]" % self._get_registry_names(self.mease.senders))

    def _get_registry_names(self, registry):
        """
        Returns functions names for a registry
        """
        return ', '.join(
            f.__name__ if not isinstance(f, tuple) else f[0].__name__
            for f in registry)

    def run(self):
        """
        Starts websocket server (blocking)
        """
        self.application.listen(self.port)

        logger.info("Websocket server listening on port {port}".format(port=self.port))

        self.ioloop.start()



# -*- coding: utf-8 -*-

__all__ = ('base', 'opener', 'closer', 'receiver', 'sender')


def method_decorator(method):
    def wrap(instance, function=None, *args, **kwargs):
        def inner_wrap(callback):
            method(instance, callback, *args, **kwargs)
        if function:
            return inner_wrap(function)
        return inner_wrap
    return wrap



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import pickle
import logging

try:
    import redis
except ImportError:
    raise ImportError('Missing backend dependency (redis)')
try:
    from toredis import Client
except ImportError:
    raise ImportError('Missing backend dependency (toredis-mease)')

from .base import BasePublisher
from .base import BaseSubscriber
from .base import BaseBackend

logger = logging.getLogger('mease.websocket_server')


class RedisBackendMixin(object):
    """
    Redis backend mixin that sets connection settings
    """
    def __init__(self, host, port, channel, *args, **kwargs):
        super(RedisBackendMixin, self).__init__(*args, **kwargs)
        self.host = host
        self.port = port
        self.channel = channel


class RedisPublisher(RedisBackendMixin, BasePublisher):
    """
    Publisher using Redis PUB
    """
    def connect(self):
        """
        Connects to publisher
        """
        self.client = redis.Redis(host=self.host, port=self.port)

    def publish(self, routing=None, *args, **kwargs):
        """
        Publishes a message
        """
        p = pickle.dumps((routing, args, kwargs), protocol=2)
        self.client.publish(self.channel, p)


class RedisSubscriber(RedisBackendMixin, BaseSubscriber):
    """
    Subscriber using Redis SUB
    """
    def connect(self):
        """
        Connects to Redis
        """
        logger.info("Connecting to Redis on {host}:{port}...".format(
            host=self.host, port=self.port))

        # Connect
        self.client = Client()
        self.client.connect(host=self.host, port=self.port)

        logger.info("Successfully connected to Redis")

        # Subscribe to channel
        self.client.subscribe(self.channel, callback=self.on_message)

        logger.info("Subscribed to [{channel}] Redis channel".format(
            channel=self.channel))

    def on_message(self, message):
        """
        Redis pubsub callback
        """
        event, channel, data = message

        if event.decode() == 'message':
            routing, args, kwargs = pickle.loads(data)
            self.dispatch_message(routing, args, kwargs)


class RedisBackend(BaseBackend):
    """
    Redis Backend using PUB/SUB
    """
    name = "Redis"
    publisher_class = RedisPublisher
    subscriber_class = RedisSubscriber

    def __init__(self, *args, **kwargs):
        super(RedisBackend, self).__init__(*args, **kwargs)

        self.host = self.settings.get('HOST', 'localhost')
        self.port = self.settings.get('PORT', 6379)
        self.channel = self.settings.get('CHANNEL', 'mease')

    def get_kwargs(self):
        """
        Returns kwargs for both publisher and subscriber classes
        """
        return {
            'host': self.host,
            'port': self.port,
            'channel': self.channel
        }

    get_publisher_kwargs = get_kwargs
    get_subscriber_kwargs = get_kwargs



# -*- coding: utf-8 -*-
import json
from concurrent.futures import ThreadPoolExecutor

from .decorators import method_decorator


class Mease(object):
    """
    Registry for mease callbacks
    """
    def __init__(self, backend_class, settings={}, max_workers=20):
        """
        Inits a registry
        """
        # Backend
        self.backend = backend_class(settings)

        self.publisher = self.backend.get_publisher()
        self.publisher.__connected = False
        self.subscriber = self.backend.get_subscriber()

        # Callbacks
        self.openers = []
        self.closers = []
        self.receivers = []
        self.senders = []

        # Executor
        self.executor = ThreadPoolExecutor(max_workers=max_workers)

    # -- Registers

    @method_decorator
    def opener(self, func):
        """
        Registers an opener function
        """
        self.openers.append(func)

    @method_decorator
    def closer(self, func):
        """
        Registers a closer function
        """
        self.closers.append(func)

    @method_decorator
    def receiver(self, func=None, json=False):
        """
        Registers a receiver function
        """
        self.receivers.append((func, json))

    @method_decorator
    def sender(self, func, routing=None):
        """
        Registers a sender function
        """
        if not isinstance(routing, list):
            routing = [routing]
        self.senders.append((func, routing))

    # -- Callers

    def call_openers(self, client, clients_list):
        """
        Calls openers callbacks
        """
        for func in self.openers:
            self.executor.submit(func, client, clients_list)

    def call_closers(self, client, clients_list):
        """
        Calls closers callbacks
        """
        for func in self.closers:
            self.executor.submit(func, client, clients_list)

    def call_receivers(self, client, clients_list, message):
        """
        Calls receivers callbacks
        """
        # Try to parse JSON
        try:
            json_message = json.loads(message)
        except ValueError:
            json_message = None

        for func, to_json in self.receivers:

            # Check if json version is available
            if to_json:
                if json_message is None:
                    continue
                msg = json_message
            else:
                msg = message

            # Call callback
            self.executor.submit(func, client, clients_list, msg)

    def call_senders(self, routing, clients_list, *args, **kwargs):
        """
        Calls senders callbacks
        """
        for func, routings in self.senders:
            if routing is None or routings is None or routing in routings:
                self.executor.submit(func, routing, clients_list, *args, **kwargs)

    # -- Publisher

    def publish(self, *args, **kwargs):
        """
        Publishes a message
        """
        # Connect to subscriber on first use
        if not self.publisher.__connected:
            self.publisher.__connected = True
            self.publisher.connect()

        self.publisher.publish(*args, **kwargs)

    # -- Websocket

    def run_websocket_server(self, port=9090, autoreload=False):
        """
        Runs Tornado websocket server (blocking)
        """
        from .server import WebSocketServer

        ws_server = WebSocketServer(self, port, autoreload)
        ws_server.run()



import json
import unittest
from .registry import Mease
from .backends.test import TestBackend


class MeaseTestCase(unittest.TestCase):

    def setUp(self):
        self.mease = Mease(TestBackend, {})
        self.ret = type("", (), {})()

    def test_opener(self):
        """
        Tests openers callbacks
        """
        client = 'a'
        clients_list = ['d', 'e', 'f']

        # Register callback
        @self.mease.opener
        def opener_func(client, clients_list):
            self.ret.client = client
            self.ret.clients_list = clients_list

        # Register other callback
        @self.mease.opener
        def otjer_opener_func(client, clients_list):
            self.ret.other_client = client
            self.ret.other_clients_list = clients_list

        # Call openers
        self.mease.call_openers(client, clients_list)
        self.mease.executor.shutdown(wait=True)

        self.assertEqual(client, self.ret.client)
        self.assertListEqual(clients_list, self.ret.clients_list)

        self.assertEqual(client, self.ret.other_client)
        self.assertListEqual(clients_list, self.ret.other_clients_list)

    def test_closer(self):
        """
        Tests closers callbacks
        """
        client = 'a'
        clients_list = ['d', 'e', 'f']

        # Register callback
        @self.mease.closer
        def closer_func(client, clients_list):
            self.ret.client = client
            self.ret.clients_list = clients_list

        # Register other callback
        @self.mease.closer
        def other_closer_func(client, clients_list):
            self.ret.other_client = client
            self.ret.other_clients_list = clients_list

        # Call closers
        self.mease.call_closers(client, clients_list)
        self.mease.executor.shutdown(wait=True)

        self.assertEqual(client, self.ret.client)
        self.assertListEqual(clients_list, self.ret.clients_list)

        self.assertEqual(client, self.ret.other_client)
        self.assertListEqual(clients_list, self.ret.other_clients_list)

    def test_receiver(self):
        """
        Tests receivers callbacks
        """
        client = 'a'
        message = '{"message": "Hello world !"}'
        clients_list = ['d', 'e', 'f']

        # Register callback
        @self.mease.receiver(json=True)
        def json_receiver_func(client, clients_list, message):
            self.ret.json_client = client
            self.ret.json_message = message
            self.ret.json_clients_list = clients_list

        # Register callback
        @self.mease.receiver
        def raw_receiver_func(client, clients_list, message):
            self.ret.raw_client = client
            self.ret.raw_message = message
            self.ret.raw_clients_list = clients_list

        # Call receivers
        self.mease.call_receivers(client, clients_list, message)
        self.mease.executor.shutdown(wait=True)

        json_message = json.loads(message)
        self.assertEqual(client, self.ret.json_client)
        self.assertEqual(json_message, self.ret.json_message)
        self.assertListEqual(clients_list, self.ret.json_clients_list)

        self.assertEqual(client, self.ret.raw_client)
        self.assertEqual(message, self.ret.raw_message)
        self.assertListEqual(clients_list, self.ret.raw_clients_list)

    def test_sender(self):
        """
        Tests senders callbacks
        """
        message = "Hello world !"
        routing = 'mease.test'

        # Register callback
        @self.mease.sender(routing='mease.test')
        def sender_func(routing, message):
            self.ret.routing = routing
            self.ret.message = message

        # This callback will not be called
        @self.mease.sender(routing='mease.other')
        def other_sender_func(routing, message):
            self.ret.other_routing = routing
            self.ret.other_message = message

        # Call senders
        self.mease.call_senders(routing, message)
        self.mease.executor.shutdown(wait=True)

        self.assertEqual(routing, self.ret.routing)
        self.assertEqual(message, self.ret.message)

        self.assertFalse(hasattr(self.ret, 'other_routing'))
        self.assertFalse(hasattr(self.ret, 'other_message'))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from .base import BasePublisher
from .base import BaseSubscriber
from .base import BaseBackend


class BasePublisher(BasePublisher):
    """
    Base publisher that handles outgoing messages
    """
    def connect(self):
        pass

    def publish(self, *args, **kwargs):
        pass


class BaseSubscriber(BaseSubscriber):
    """
    Base subscriber class that handles incoming messages
    """
    def connect(self):
        pass


class TestBackend(BaseBackend):
    """
    Base backend with a publisher and a subscriber class
    """
    publisher_class = BasePublisher
    subscriber_class = BaseSubscriber



# -*- coding: utf-8 -*-
import logging

logger = logging.getLogger('mease.websocket_server')


class BasePublisher(object):
    """
    Base publisher that handles outgoing messages
    """
    def connect(self):
        raise NotImplementedError(
            "You need to implement your `connect` method for your publisher")

    def publish(self, *args, **kwargs):
        raise NotImplementedError(
            "You need to implement your `publish` method for your publisher")


class BaseSubscriber(object):
    """
    Base subscriber class that handles incoming messages
    """
    def connect(self):
        raise NotImplementedError(
            "You need to implement your `connect` method for your subscriber")

    def dispatch_message(self, routing, args, kwargs):
        """
        Calls callback functions
        """
        logger.debug("Backend message on [{routing}] route : {args} {kwargs}".format(
            routing=routing, args=args, kwargs=kwargs))

        self.application._mease.call_senders(
            routing, self.application.clients, *args, **kwargs)


class BaseBackend(object):
    """
    Base backend with a publisher and a subscriber class
    """
    publisher_class = BasePublisher
    subscriber_class = BaseSubscriber

    def __init__(self, settings):
        self.settings = settings

    def get_publisher_kwargs(self):
        """
        Additional kwargs for publisher instance
        """
        return {}

    def get_publisher(self):
        """
        Returns a publisher instance
        """
        return self.publisher_class(**self.get_publisher_kwargs())

    def get_subscriber_kwargs(self):
        """
        Additional kwargs for subscriber instance
        """
        return {}

    def get_subscriber(self):
        """
        Returns a subscriber instance
        """
        return self.subscriber_class(**self.get_subscriber_kwargs())



# -*- coding: utf-8 -*-
import pickle
import logging

try:
    import pika
    from pika.adapters.tornado_connection import TornadoConnection
except ImportError:
    raise ImportError('Missing backend dependency (pika)')

from .base import BasePublisher
from .base import BaseSubscriber
from .base import BaseBackend


logger = logging.getLogger('mease.websocket_server')


class RabbitMQBackendMixin(object):
    """
    RabbitMQ backend mixin that stores connection settings
    """
    def __init__(self, broker_url, queue, *args, **kwargs):
        super(RabbitMQBackendMixin, self).__init__(*args, **kwargs)

        self.broker_url = broker_url
        self.queue = queue

        self.params = pika.URLParameters(self.broker_url)


class RabbitMQPublisher(RabbitMQBackendMixin, BasePublisher):
    """
    RabbitMQ publisher
    """
    def connect(self):
        self.client = pika.BlockingConnection(self.params)
        self.channel = self.client.channel()

    def publish(self, routing=None, *args, **kwargs):
        """
        Publishes a message
        """
        p = pickle.dumps((routing, args, kwargs), protocol=2)
        self.channel.basic_publish(exchange='', routing_key=self.queue, body=p)


class RabbitMQSubscriber(RabbitMQBackendMixin, BaseSubscriber):
    """
    RabbitMQ subscriber
    """
    def __init__(self, *args, **kwargs):
        super(RabbitMQSubscriber, self).__init__(*args, **kwargs)

        self.connection = None
        self.consumer_tag = None
        self.channel = None

    def __connect(self):
        """
        Connects to RabbitMQ
        """
        logger.info("Connecting to RabbitMQ on {broker_url}...".format(
            broker_url=self.broker_url))

        return TornadoConnection(
            self.params, on_open_callback=self.on_connected, stop_ioloop_on_close=False)

    def connect(self):
        self.connection = self.__connect()

    def on_connected(self, connection):
        """
        Opens a channel after connection
        """
        logger.info("Successfully connected to RabbitMQ")

        self.connection.add_on_close_callback(self.on_closed)
        self.connection.channel(self.on_channel_open)

    def on_channel_open(self, channel):
        """
        Declares a queue after channel connection
        """
        self.channel = channel
        self.channel.queue_declare(
            queue=self.queue,
            durable=True,
            exclusive=False,
            auto_delete=False,
            callback=self.on_queue_declared)

    def on_queue_declared(self, frame):
        """
        Registers a consumer on queue
        """
        logger.info("Subscribed to [{queue}] RabbitMQ queue".format(
            queue=self.queue))

        self.consumer_tag = self.channel.basic_consume(
            self.handle_delivery, queue=self.queue)

    def on_closed(self, connection, reply_code, reply_text):
        """
        On close
        """
        self.channel = None
        self.connection.add_timeout(2, self.reconnect)

    def reconnect(self):
        """
        Reconnects to amqp
        """
        logger.info(
            "Lost connection to RabbitMQ... Reconnecting to {broker_url}...".format(
                broker_url=self.broker_url))

        self.connection = self.connect()

    def handle_delivery(self, channel, method, header, body):
        """
        Handles message delivery
        """
        routing, args, kwargs = pickle.loads(body)

        # Dispatch to callbacks
        self.dispatch_message(routing, args, kwargs)

        self.acknowledge_message(method.delivery_tag)

    def acknowledge_message(self, delivery_tag):
        """
        Acknowledge message
        """
        self.channel.basic_ack(delivery_tag)


class RabbitMQBackend(BaseBackend):
    name = "RabbitMQ"
    publisher_class = RabbitMQPublisher
    subscriber_class = RabbitMQSubscriber

    def __init__(self, *args, **kwargs):
        super(RabbitMQBackend, self).__init__(*args, **kwargs)

        self.broker_url = self.settings.get(
            'BROKER_URL', 'amqp://guest:guest@localhost:5672//')
        self.queue = self.settings.get('QUEUE', 'mease')

    def get_kwargs(self):
        """
        Returns kwargs for both publisher and subscriber classes
        """
        return {
            'broker_url': self.broker_url,
            'queue': self.queue
        }

    get_publisher_kwargs = get_kwargs
    get_subscriber_kwargs = get_kwargs



# -*- coding: utf-8 -*-
from functools import wraps


def passes_test(perm_func):
    def decored(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not perm_func(*args, **kwargs):
                return
            func(*args, **kwargs)
        return wrapper
    return decored



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

requires = [
    'tornado>=3.2',
]

if sys.version_info[0] == 2:
    requires.append('futures')

setup(
    name='mease',
    version='0.1.1',
    description="Mease: Tornado websocket server with an easy callback registry",
    url="https://github.com/florianpaquet/mease",
    author="Florian PAQUET",
    author_email="contact@florianpaquet.com",
    long_description=read('README.rst'),
    license='MIT',
    test_suite='mease.tests',
    packages=[
        'mease',
        'mease.backends',
    ],
    install_requires=requires,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ])



# -*- coding: utf-8 -*-


class SettingsMixin(object):
    def get_setting(self, settings, parent, name, default=None):
        """
        Returns a setting from a settings dict
        """
        root = settings

        if parent:
            root = settings.get(parent, {})

        return root.get(name, default)
